###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.7.7.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""
This module contains a class, MerlinSpec, which holds the unchanged
data from the Merlin specification file.
To see examples of yaml specifications, run `merlin example`.
"""
import logging
import os
from io import StringIO

import yaml
from maestrowf.datastructures import YAMLSpecification

from merlin.spec import all_keys, defaults


LOG = logging.getLogger(__name__)


class MerlinSpec(YAMLSpecification):
    """
    This class represents the logic for parsing the Merlin yaml
    specification.

    Example spec_file contents:

    --spec_file.yaml--
    ...
    merlin:
        resources:
            task_server: celery
        samples:
            generate:
                cmd: python make_samples.py -outfile=$(OUTPUT_PATH)/merlin_info/samples.npy
            file: $(OUTPUT_PATH)/merlin_info/samples.npy
            column_labels: [X0, X1]
    """

    def __init__(self):
        super(MerlinSpec, self).__init__()

    @property
    def yaml_sections(self):
        """
        Returns a nested dictionary of all sections of the specification
        as used in a yaml spec.
        """
        return {
            "description": self.description,
            "batch": self.batch,
            "env": self.environment,
            "study": self.study,
            "global.parameters": self.globals,
            "merlin": self.merlin,
        }

    @property
    def sections(self):
        """
        Returns a nested dictionary of all sections of the specification
        as referenced by Maestro's YAMLSpecification class.
        """
        return {
            "description": self.description,
            "batch": self.batch,
            "environment": self.environment,
            "study": self.study,
            "globals": self.globals,
            "merlin": self.merlin,
        }

    @classmethod
    def load_specification(cls, filepath, suppress_warning=True):
        spec = super(MerlinSpec, cls).load_specification(filepath)
        with open(filepath, "r") as f:
            spec.merlin = MerlinSpec.load_merlin_block(f)
        spec.specroot = os.path.dirname(spec.path)
        spec.process_spec_defaults()
        if not suppress_warning:
            spec.warn_unrecognized_keys()
        return spec

    @classmethod
    def load_spec_from_string(cls, string):
        spec = super(MerlinSpec, cls).load_specification_from_stream(StringIO(string))
        spec.merlin = MerlinSpec.load_merlin_block(StringIO(string))
        spec.specroot = None
        spec.process_spec_defaults()
        return spec

    @staticmethod
    def load_merlin_block(stream):
        try:
            merlin_block = yaml.safe_load(stream)["merlin"]
        except KeyError:
            merlin_block = {}
            LOG.warning(
                f"Workflow specification missing \n "
                f"encouraged 'merlin' section! Run 'merlin example' for examples.\n"
                f"Using default configuration with no sampling."
            )
        return merlin_block

    def process_spec_defaults(self):
        for name, section in self.sections.items():
            if section is None:
                setattr(self, name, {})

        # fill in missing batch section defaults
        MerlinSpec.fill_missing_defaults(self.batch, defaults.BATCH["batch"])

        # fill in missing env section defaults
        MerlinSpec.fill_missing_defaults(self.environment, defaults.ENV["env"])

        # fill in missing global parameter section defaults
        MerlinSpec.fill_missing_defaults(
            self.globals, defaults.PARAMETER["global.parameters"]
        )

        # fill in missing step section defaults within 'run'
        defaults.STUDY_STEP_RUN["shell"] = self.batch["shell"]
        for step in self.study:
            MerlinSpec.fill_missing_defaults(step["run"], defaults.STUDY_STEP_RUN)

        # fill in missing merlin section defaults
        MerlinSpec.fill_missing_defaults(self.merlin, defaults.MERLIN["merlin"])
        if self.merlin["resources"]["workers"] is None:
            self.merlin["resources"]["workers"] = {"default_worker": defaults.WORKER}
        else:
            for worker, vals in self.merlin["resources"]["workers"].items():
                MerlinSpec.fill_missing_defaults(vals, defaults.WORKER)
        if self.merlin["samples"] is not None:
            MerlinSpec.fill_missing_defaults(self.merlin["samples"], defaults.SAMPLES)

    @staticmethod
    def fill_missing_defaults(object_to_update, default_dict):
        """
        Merge keys and values from a dictionary of defaults
        into a parallel object that may be missing attributes.
        Only adds missing attributes to object; does not overwrite
        existing ones.
        """

        def recurse(result, defaults):
            if not isinstance(defaults, dict):
                return
            for key, val in defaults.items():
                if (key not in result) or (
                    (result[key] is None) and (defaults[key] is not None)
                ):
                    result[key] = val
                else:
                    recurse(result[key], val)

        recurse(object_to_update, default_dict)

    def warn_unrecognized_keys(self):
        # check description
        MerlinSpec.check_section("description", self.description, all_keys.DESCRIPTION)

        # check batch
        MerlinSpec.check_section("batch", self.batch, all_keys.BATCH)

        # check env
        MerlinSpec.check_section("env", self.environment, all_keys.ENV)

        # check parameters
        for param, contents in self.globals.items():
            MerlinSpec.check_section("global.parameters", contents, all_keys.PARAMETER)

        # check steps
        for step in self.study:
            MerlinSpec.check_section(step["name"], step, all_keys.STUDY_STEP)
            MerlinSpec.check_section(
                step["name"] + ".run", step["run"], all_keys.STUDY_STEP_RUN
            )

        # check merlin
        MerlinSpec.check_section("merlin", self.merlin, all_keys.MERLIN)
        MerlinSpec.check_section(
            "merlin.resources", self.merlin["resources"], all_keys.MERLIN_RESOURCES
        )
        for worker, contents in self.merlin["resources"]["workers"].items():
            MerlinSpec.check_section(
                "merlin.resources.workers " + worker, contents, all_keys.WORKER
            )
        if self.merlin["samples"]:
            MerlinSpec.check_section(
                "merlin.samples", self.merlin["samples"], all_keys.SAMPLES
            )

    @staticmethod
    def check_section(section_name, section, all_keys):
        diff = set(section.keys()).difference(all_keys)
        for extra in diff:
            LOG.warn(
                f"Unrecognized key '{extra}' found in spec section '{section_name}'."
            )

    def dump(self):
        """
        Dump this MerlinSpec to a pretty yaml string.
        """
        tab = "   "
        list_offset = "  "
        from copy import deepcopy

        def dict_to_yaml(obj, string, key_stack, newline=True):
            if obj is None:
                return ""
            lvl = len(key_stack) - 1
            if isinstance(obj, str):
                split = obj.splitlines()
                if len(split) > 1:
                    obj = "|\n" + tab * (lvl + 1) + ("\n" + tab * (lvl + 1)).join(split)
                return obj
            if isinstance(obj, bool):
                return str(obj).lower()
            if (not isinstance(obj, list)) and (not isinstance(obj, dict)):
                return obj
            if isinstance(obj, list):
                n = len(obj)
                use_hyphens = key_stack[-1] in ["paths", "sources", "git", "study"]
                if not use_hyphens:
                    string += "["
                else:
                    string += "\n"
                for i, elem in enumerate(obj):
                    key_stack = deepcopy(key_stack)
                    key_stack.append("elem")
                    if use_hyphens:
                        string += (
                            (lvl + 1) * tab
                            + "- "
                            + str(dict_to_yaml(elem, "", key_stack))
                            + "\n"
                        )
                    else:
                        string += str(
                            dict_to_yaml(elem, "", key_stack, newline=(i != 0))
                        )
                        if n > 1 and i != len(obj) - 1:
                            string += ", "
                    key_stack.pop()
                if not use_hyphens:
                    string += "]"
            if isinstance(obj, dict):
                if len(key_stack) > 0 and key_stack[-1] != "elem":
                    string += "\n"
                i = 0
                for k, v in obj.items():
                    key_stack = deepcopy(key_stack)
                    key_stack.append(k)
                    if len(key_stack) > 1 and key_stack[-2] == "elem" and i == 0:
                        # string += (tab * (lvl - 1))
                        string += ""
                    elif "elem" in key_stack:
                        string += list_offset + (tab * lvl)
                    else:
                        string += tab * (lvl + 1)
                    string += str(k) + ": " + str(dict_to_yaml(v, "", key_stack)) + "\n"
                    key_stack.pop()
                    i += 1
            return string

        result = dict_to_yaml(self.yaml_sections, "", [])
        while "\n\n\n" in result:
            result = result.replace("\n\n\n", "\n\n")
        try:
            yaml.safe_load(result)
        except BaseException as e:
            raise ValueError(f"Error parsing provenance spec:\n{e}")
        return result

    def get_task_queues(self):
        """Returns a dictionary of steps and their corresponding task queues."""
        steps = self.get_study_steps()
        queues = {}
        for step in steps:
            if "task_queue" in step.run:
                queues[step.name] = step.run["task_queue"]
        return queues

    def get_queue_list(self, steps):
        """
        Return a sorted list of queues corresponding to spec steps

        param steps: a list of step names or 'all'
        """
        queues = self.get_task_queues()
        if steps[0] == "all":
            task_queues = queues.values()
        else:
            try:
                if isinstance(steps, list):
                    task_queues = [queues[step] for step in steps]
                else:
                    task_queues = [queues[steps]]
            except KeyError:
                nl = "\n"
                LOG.error(
                    f"Invalid steps '{steps}'! Try one of these (or 'all'):\n{nl.join(queues.keys())}"
                )
                raise
        return sorted(set(task_queues))

    def make_queue_string(self, steps):
        """
        Return a unique queue string for the steps

        param steps: a list of step names
        """
        return ",".join(set(self.get_queue_list(steps)))

    def get_worker_names(self):
        result = []
        for worker in self.merlin["resources"]["workers"]:
            result.append(worker)
        return result



###############################################################################
# Copyright (c) 2019, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by the Merlin dev team, listed in the CONTRIBUTORS file.
# <merlin@llnl.gov>
#
# LLNL-CODE-797170
# All rights reserved.
# This file is part of Merlin, Version: 1.7.7.
#
# For details, see https://github.com/LLNL/merlin.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
###############################################################################

"""Updated celery configuration."""
from __future__ import absolute_import, print_function

import logging
import os

import billiard
import psutil
from celery import Celery
from celery.signals import worker_process_init

import merlin.common.security.encrypt_backend_traffic
from merlin.config import broker, celeryconfig, results_backend
from merlin.config.configfile import CONFIG
from merlin.router import route_for_task
from merlin.utils import nested_namespace_to_dicts


LOG = logging.getLogger(__name__)

merlin.common.security.encrypt_backend_traffic.set_backend_funcs()

broker_ssl = True
results_ssl = False
try:
    BROKER_URI = broker.get_connection_string()
    LOG.debug(f"broker: {broker.get_connection_string(include_password=False)}")
    broker_ssl = broker.get_ssl_config()
    LOG.debug(f"broker_ssl = {broker_ssl}")
    RESULTS_BACKEND_URI = results_backend.get_connection_string()
    results_ssl = results_backend.get_ssl_config(celery_check=True)
    LOG.debug(
        f"results: {results_backend.get_connection_string(include_password=False)}"
    )
    LOG.debug(f"results: redis_backed_use_ssl = {results_ssl}")
except ValueError:
    # These variables won't be set if running with '--local'.
    BROKER_URI = None
    RESULTS_BACKEND_URI = None

# initialize app with essential properties
app = Celery(
    "merlin",
    broker=BROKER_URI,
    backend=RESULTS_BACKEND_URI,
    broker_use_ssl=broker_ssl,
    redis_backend_use_ssl=results_ssl,
    task_routes=(route_for_task,),
)

# load merlin config defaults
app.conf.update(**celeryconfig.DICT)

# load config overrides from app.yaml
if (
    (not hasattr(CONFIG.celery, "override"))
    or (CONFIG.celery.override is None)
    or (len(nested_namespace_to_dicts(CONFIG.celery.override)) == 0)
):
    LOG.debug("Skipping celery config override; 'celery.override' field is empty.")
else:
    override_dict = nested_namespace_to_dicts(CONFIG.celery.override)
    override_str = ""
    i = 0
    for k, v in override_dict.items():
        if k not in str(app.conf.__dict__):
            raise ValueError(f"'{k}' is not a celery configuration.")
        override_str += f"\t{k}:\t{v}"
        if i != len(override_dict) - 1:
            override_str += "\n"
        i += 1
    LOG.info(
        f"Overriding default celery config with 'celery.override' in 'app.yaml':\n{override_str}"
    )
    app.conf.update(**override_dict)

# auto-discover tasks
app.autodiscover_tasks(["merlin.common"])


@worker_process_init.connect()
def setup(**kwargs):
    """
    Set affinity for the worker on startup (works on toss3 nodes)

    :param `**kwargs`: keyword arguments
    """
    if "CELERY_AFFINITY" in os.environ and int(os.environ["CELERY_AFFINITY"]) > 1:
        # Number of cpus between workers.
        cpu_skip = int(os.environ["CELERY_AFFINITY"])
        npu = psutil.cpu_count()
        p = psutil.Process()
        current = billiard.current_process()
        prefork_id = current._identity[0] - 1  # range 0:nworkers-1
        cpu_slot = (prefork_id * cpu_skip) % npu
        p.cpu_affinity(list(range(cpu_slot, cpu_slot + cpu_skip)))



import argparse

import numpy as np
from joblib import dump, load
from sklearn.ensemble import RandomForestRegressor


parser = argparse.ArgumentParser("Learn surrogate model form simulation")
parser.add_argument(
    "-collector_dir",
    help="Collector directory (.npz file), usually '$(collector.workspace)'",
)
args = parser.parse_args()

collector_dir = args.collector_dir
current_iter_npz = np.load(f"{collector_dir}/current_results.npz", allow_pickle=True)
current_iter_data = current_iter_npz["arr_0"].item()

try:
    prev_iter_npz = np.load("all_iter_results.npz", allow_pickle=True)
    prev_iter_data = prev_iter_npz["arr_0"].item()
    data = dict(prev_iter_data, **current_iter_data)
except:
    data = current_iter_data

X = []
y = []

for i in data.keys():
    X.append(data[i]["Inputs"])
    y.append(data[i]["Outputs"])

X = np.array(X)
y = np.array(y)

surrogate = RandomForestRegressor(max_depth=4, random_state=0, n_estimators=100)
surrogate.fit(X, y)

dump(surrogate, "surrogate.joblib")
np.savez("all_iter_results.npz", data)



from maestrowf.datastructures.core import ParameterGenerator


def get_custom_generator(env, **kwargs):
    p_gen = ParameterGenerator()
    params = {
        "X2": {"values": [1 / i for i in range(3, 6)], "label": "X2.%%"},
        "N_NEW": {"values": [2 ** i for i in range(1, 4)], "label": "N_NEW.%%"},
    }

    for key, value in params.items():
        p_gen.add_parameter(key, value["values"], value["label"])

    return p_gen



import logging
from copy import deepcopy


LOG = logging.getLogger(__name__)


def error_override_vars(override_vars, spec_filepath):
    """
    Warn user if any given variable name isn't found in the original spec file.
    """
    if override_vars is None:
        return
    original_text = open(spec_filepath, "r").read()
    for variable in override_vars.keys():
        if variable not in original_text:
            raise ValueError(
                f"Command line override variable '{variable}' not found in spec file '{spec_filepath}'."
            )


def replace_override_vars(env, override_vars):
    if override_vars is None:
        return env
    result = deepcopy(env)
    for key, val in env.items():
        for var_name, var_val in override_vars.items():
            if var_name in val:
                result[key][var_name] = var_val
    return result



import argparse

import numpy as np


# argument parsing
parser = argparse.ArgumentParser(description="Make some samples (names of people).")
parser.add_argument(
    "--number", type=int, action="store", help="the number of samples you want to make"
)
parser.add_argument(
    "--filepath", type=str, default="samples_file.npy", help="output file"
)
args = parser.parse_args()

# sample making
result = np.random.random((args.number, 1))

np.save(args.filepath, result)



import argparse
import json

import numpy as np


def N_Rosenbrock(X):
    X = X.T
    total = 0
    for i in range(X.shape[0] - 1):
        total += 100 * ((X[i + 1] - X[i] ** 2) ** 2) + (1 - X[i]) ** 2

    return total


parser = argparse.ArgumentParser("Generate some samples!")
parser.add_argument("-ID")
parser.add_argument("-inputs", nargs="+")
args = parser.parse_args()

run_id = args.ID
inputs = args.inputs
inputs = np.array(inputs).astype(np.float)
outputs = N_Rosenbrock(inputs)

results = {run_id: {"Inputs": inputs.tolist(), "Outputs": outputs.tolist()}}

json.dump(results, open("simulation_results.json", "w"), indent=4)



"""
Default celery configuration for merlin
"""

from merlin.log_formatter import FORMATS


DICT = {
    "task_serializer": "pickle",
    "accept_content": ["pickle"],
    "result_serializer": "pickle",
    "task_acks_late": True,
    "task_reject_on_worker_lost": True,
    "task_publish_retry_policy": {
        "interval_start": 10,
        "interval_step": 10,
        "interval_max": 60,
    },
    "redis_max_connections": 100000,
    "broker_transport_options": {
        "visibility_timeout": 60 * 60 * 24,
        "max_connections": 100,
    },
    "broker_pool_limit": 0,
    "task_default_queue": "merlin",
    "worker_log_color": True,
    "worker_log_format": FORMATS["DEFAULT"],
    "worker_task_log_format": FORMATS["WORKER"],
}



import argparse

import matplotlib.pyplot as plt
import numpy as np
from joblib import dump, load


plt.style.use("seaborn-white")

parser = argparse.ArgumentParser("Learn surrogate model form simulation")
parser.add_argument(
    "-study_dir", help="The study directory, usually '$(MERLIN_WORKSPACE)'"
)
args = parser.parse_args()

study_dir = args.study_dir
npz_path = f"{study_dir}/learner/all_iter_results.npz"
learner_path = f"{study_dir}/learner/surrogate.joblib"
new_samples_path = f"{study_dir}/pick_new_inputs/new_samples.npy"
new_exploit_samples_path = f"{study_dir}/pick_new_inputs/new_exploit_samples.npy"
new_explore_samples_path = f"{study_dir}/pick_new_inputs/new_explore_samples.npy"
new_explore_star_samples_path = (
    f"{study_dir}/pick_new_inputs/new_explore_star_samples.npy"
)
optimum_path = f"{study_dir}/optimizer/optimum.npy"
old_best_path = f"{study_dir}/optimizer/old_best.npy"

from_file = np.load(npz_path, allow_pickle=True)

data = from_file["arr_0"].item()

X = []
y = []

for i in data.keys():
    X.append(data[i]["Inputs"])
    y.append(data[i]["Outputs"])

existing_X = np.array(X)
existing_y = np.array(y)

surrogate = load(learner_path)

new_samples = np.load(new_samples_path)
new_exploit_samples = np.load(new_exploit_samples_path)
new_explore_samples = np.load(new_explore_samples_path)
new_explore_star_samples = np.load(new_explore_star_samples_path)
optimum = np.load(optimum_path)
old_best = np.load(old_best_path)


def Rosenbrock_mesh():
    X_mesh_plot = np.array([np.linspace(-2, 2, n_points), np.linspace(-1, 3, n_points)])
    X_mesh = np.meshgrid(X_mesh_plot[0], X_mesh_plot[1])

    Z_mesh = (1 - X_mesh[0]) ** 2 + 100 * (X_mesh[1] - X_mesh[0] ** 2) ** 2

    return X_mesh, Z_mesh


# Script for N_dim Rosenbrock function
n_points = 250

X_mesh, Z_mesh = Rosenbrock_mesh()

Z_pred = surrogate.predict(np.c_[X_mesh[0].ravel(), X_mesh[1].ravel()])
Z_pred = Z_pred.reshape(X_mesh[0].shape)

# Surface plot
plt.rcParams.update({"font.size": 15})
fig = plt.figure(figsize=(30, 45))

ax = fig.add_subplot(3, 2, 1, projection="3d")
ax.plot_surface(
    X_mesh[0],
    X_mesh[1],
    Z_mesh,
    rstride=5,
    cstride=5,
    cmap="jet",
    alpha=0.4,
    edgecolor="none",
)
ax.scatter(existing_X[:, 0], existing_X[:, 1], existing_y, marker="x")
ax.view_init(45, 45)
ax.set_xlabel("DIM_1")
ax.set_ylabel("DIM_2")
ax.set_title("Rosenbrock Function")

ax = fig.add_subplot(3, 2, 2, projection="3d")
ax.plot_surface(
    X_mesh[0],
    X_mesh[1],
    Z_pred,
    rstride=5,
    cstride=5,
    cmap="jet",
    alpha=0.4,
    edgecolor="none",
)
ax.scatter(existing_X[:, 0], existing_X[:, 1], existing_y, marker="x")
ax.view_init(45, 45)
ax.set_xlabel("DIM_1")
ax.set_ylabel("DIM_2")
ax.set_title("Surrogate Function")

ax = fig.add_subplot(3, 2, 3, projection="3d")
ax.plot_surface(
    X_mesh[0],
    X_mesh[1],
    np.clip(Z_mesh, -100, 100),
    rstride=5,
    cstride=5,
    cmap="jet",
    alpha=0.4,
    edgecolor="none",
)
ax.scatter(
    existing_X[:, 0], existing_X[:, 1], np.clip(existing_y, -100, 100), marker="x"
)
ax.view_init(45, 45)
ax.set_xlabel("DIM_1")
ax.set_ylabel("DIM_2")
ax.set_title("Rosenbrock Function clipped")

ax = fig.add_subplot(3, 2, 4, projection="3d")
ax.plot_surface(
    X_mesh[0],
    X_mesh[1],
    np.clip(Z_pred, -100, 100),
    rstride=5,
    cstride=5,
    cmap="jet",
    alpha=0.4,
    edgecolor="none",
)
ax.scatter(
    existing_X[:, 0], existing_X[:, 1], np.clip(existing_y, -100, 100), marker="x"
)
ax.view_init(45, 45)
ax.set_xlabel("DIM_1")
ax.set_ylabel("DIM_2")
ax.set_title("Surrogate Function clipped")

ax = fig.add_subplot(3, 2, 5)
ax.scatter(existing_X[:, 0], existing_X[:, 1], label="Existing Inputs")
ax.scatter(new_samples[:, 0], new_samples[:, 1], label="Suggested Inputs")
ax.annotate("Predicted Optimum", xy=optimum, xytext=(1, 0), arrowprops=dict(width=0.01))
ax.annotate(
    "Current Best", xy=old_best, xytext=(-1.5, 1.5), arrowprops=dict(width=0.01)
)
ax.annotate("Actual Minimum", xy=(1, 1), xytext=(1.5, 2.5), arrowprops=dict(width=0.01))

ax.set_xlabel("DIM_1")
ax.set_ylabel("DIM_2")
ax.set_title("Existing Inputs vs Suggested Inputs")
ax.legend()
ax.grid()

ax = fig.add_subplot(3, 2, 6)
ax.scatter(
    new_exploit_samples[:, 0], new_exploit_samples[:, 1], label="Exploit Samples"
)
ax.scatter(
    new_explore_samples[:, 0], new_explore_samples[:, 1], label="Explore Samples"
)
ax.scatter(
    new_explore_star_samples[:, 0],
    new_explore_star_samples[:, 1],
    label="Explore_star Samples",
)
ax.annotate("Predicted Optimum", xy=optimum)
ax.annotate("Current Best", xy=old_best)

ax.set_xlabel("DIM_1")
ax.set_ylabel("DIM_2")
ax.set_title("Suggested Inputs zoomed in")
ax.legend()
ax.grid()

plt.savefig("results.png")
